# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

# This is a "shim" setup.py file that delegates all configuration
# to the pyproject.toml file. This is the recommended approach for
# projects that need to maintain a setup.py for compatibility while
# adopting modern packaging standards.
#
# All metadata, dependencies, and package data are defined in pyproject.toml.
# This setup() call is intentionally left empty.
setup()